import unittest
from unittest.mock import Mock, patch

from flotorch_core.inferencer.inferencer_provider_factory import InferencerProviderFactory
from flotorch_core.inferencer.gateway_inferencer import GatewayInferencer
from flotorch_core.inferencer.inferencer import DEFAULT_SYSTEM_PROMPT


class TestGatewayInferencerFactory(unittest.TestCase):
    """Test GatewayInferencer creation through InferencerProviderFactory"""
    
    def test_create_gateway_inferencer_through_factory(self):
        """Test creating GatewayInferencer using the factory with gateway_enabled=True"""
        inferencer = InferencerProviderFactory.create_inferencer_provider(
            gateway_enabled=True,
            base_url="",
            api_key="",
            service="",
            model_id="flotorch/haiku-long",
            region="",
            arn_role=""
        )
        
        metadata, response = inferencer.generate_text("What is the capital of France?", [])

        self.assertIsInstance(inferencer, GatewayInferencer)
        self.assertEqual(inferencer.model_id, "flotorch/haiku-long")
        self.assertEqual(inferencer.api_key, "")
        self.assertEqual(inferencer.base_url, "")
        self.assertIsNone(inferencer.n_shot_prompt_guide_obj)
        self.assertIsNotNone(response)
        self.assertIsNotNone(metadata)


if __name__ == '__main__':
    unittest.main()



"""
This class is responsible for embedding the text using the Gateway model.
"""
from typing import Dict
from openai import OpenAI
from flotorch_core.embedding.embedding import BaseEmbedding
from flotorch_core.chunking.chunking import Chunk
from flotorch_core.embedding.embedding import Embeddings, EmbeddingMetadata
from flotorch_core.embedding.embedding_registry import register


@register("gateway")
class GatewayEmbedding(BaseEmbedding):
    """
    Initializes the GatewayEmbedding class.
    :param model_id: The model id of the Gateway model.
    :param base_url: The base url of the console.
    :param api_key: The api key of the console.
    :param headers: The headers of the console.
    :param dimensions: The dimensions of the embedding.
    :param normalize: Normalize the embeddings.
    """
    def __init__(
        self,
        model_id: str,
        base_url: str,
        api_key: str,
        headers: Dict[str, str] = None,
        dimensions: int = 256,
        normalize: bool = True,
    ):
        super().__init__(model_id, None, dimensions, normalize)
        self.base_url = base_url
        self.api_key = api_key
        self.headers = headers or {}
        self.client = OpenAI(
            api_key=self.api_key, base_url=self.base_url, default_headers=self.headers
        )

    def _prepare_chunk(self, chunk: Chunk) -> Dict:
        return {"input": chunk.data}

    def embed(self, chunk: Chunk) -> Embeddings:
        response = self.client.embeddings.create(input=chunk.data, model=self.model_id)
        metadata = EmbeddingMetadata(
            input_tokens=response.usage.total_tokens, latency_ms=0.0
        )
        return Embeddings(
            embeddings=response.data[0].embedding, metadata=metadata, text=chunk.data
        )



from typing import List, Dict, Any, Optional,Type,Union
from deepeval import evaluate
from deepeval.evaluate import AsyncConfig
from deepeval.test_case import LLMTestCase
from flotorch_core.evaluator.base_evaluator import BaseEvaluator
from flotorch_core.evaluator.evaluation_item import EvaluationItem
from flotorch_core.evaluator.metrics.deepeval_metrics.deepeval_metrics import DeepEvalEvaluationMetrics
from deepeval.models.base_model import DeepEvalBaseLLM
from flotorch_core.inferencer.inferencer import BaseInferencer
from pydantic import BaseModel
from flotorch_core.evaluator.metrics.metrics_keys import MetricKey
from deepeval.models.llms.utils import trim_and_load_json
from flotorch_core.logger.global_logger import get_logger
from deepeval.evaluate import ErrorConfig
from tenacity import retry, wait_exponential_jitter, retry_if_exception_type, stop_after_attempt
import json
logger = get_logger()


class DeepEvalEvaluator(BaseEvaluator):
    """
    Evaluator that uses DeepEval metrics to evaluate LLM outputs with optional custom metrics
    and support for asynchronous evaluation.

    Initializes with an LLM inferencer and allows configuration of custom metrics, asynchronous execution,
    concurrency limits, and optional metric-specific arguments.
    Args:
        evaluator_llm : The LLM inferencer used for evaluation.
        custom_metrics :A list of additional metric instances to include in evaluation beyond the default DeepEval metrics registry.
        async_run :Whether to run evaluation asynchronously.If True, evaluation can run concurrently up to `max_concurrent` tasks.
        max_concurrent : Maximum number of concurrent asynchronous evaluation tasks to run.    
        metric_args :Optional dictionary specifying per-metric configuration arguments.
    Example:
        metric_args = {
            "contextual_recall": {
                "threshold": 0.6
            },
            "hallucination": {
                "threshold": 0.4
            }
        }
    """

    def __init__(
        self,
        evaluator_llm: BaseInferencer,
        custom_metrics: Optional[List[Any]] = None,
        async_run: bool = False,
        max_concurrent: int = 1, 
        metric_args: Optional[
            Dict[Union[str, MetricKey], Dict[str, Union[str, float, int]]]
        ] = None

    ):
        class FloTorchLLMWrapper(DeepEvalBaseLLM):
            def __init__(self, inference_llm: BaseInferencer, *args, **kwargs):
                self.inference_llm = inference_llm
                super().__init__(*args, **kwargs)

            def get_model_name(self) -> str:
                """
                Returns the model ID of the underlying inference LLM.
                """
                return self.inference_llm.model_id

            def generate(self, prompt: str, schema: Optional[Type[BaseModel]] = None) -> str: 
                """
                Generates a response for a prompt and validates it against a schema if provided.
                """
                client = self.load_model()
                _, completion = client.generate_text(prompt, None)
                return self.schema_validation(completion, schema)

            async def a_generate(self, prompt: str, schema: Optional[Type[BaseModel]] = None) -> str: 
                """
                Asynchronously generates a response for a prompt and validates it against a schema if provided.
                """
                client = self.load_model()
                _, completion = await client.generate_text(prompt, None)                
                return self.schema_validation(completion, schema)

            def load_model(self):
                """
                Loads and returns the inference LLM client.
                """
                return self.inference_llm
            
           
            @retry(
                wait=wait_exponential_jitter(initial=1, exp_base=2, jitter=2, max=10),
                retry=retry_if_exception_type(ValueError),
                stop=stop_after_attempt(3)
            )
            def schema_validation(self, completion: str, schema: Optional[Type[BaseModel]] = None) -> str:
                try:
                    if schema:
                        json_output = self.trim_json(completion)
                        parsed_output = json.loads(json_output)
                        return schema.model_validate(parsed_output)
                    else:
                        return completion
                except ValueError as ve:
                    raise ve  
                except Exception as e:
                    logger.error(f"Schema validation error due to {e}.")
                    return completion

                
            def llm_fix_json_prompt(self, bad_json: str) -> str:
                return f"""The following is a malformed JSON (possibly incomplete or with syntax issues). Fix it so that it becomes **valid JSON**.
                        Instructions:
                        - Do **NOT** include Markdown formatting (no triple backticks, no ```json).
                        - Do **NOT** add or invent any new keys or values.
                        - Only fix unclosed strings, arrays, or braces.
                        - Do **NOT** add commas or fields that were not originally present.
                        - **Remove any trailing commas** at the end of JSON objects or arrays.
                        - **Ensure all property names and string values are enclosed in double quotes**.
                        - Preserve the original structure and values.
                        - If a list like "truths" or "verdicts" or "statements" or "reason" seems incomplete, just close it properly.
                        - Output **only** valid, raw JSON. No explanation, no surrounding text, no markdown.

                        Malformed JSON to fix:
                        {json.dumps(bad_json)}
                        """
            def fix_common_truncation(self,json_str: str) -> str:
                if not json_str.endswith(']') and not json_str.endswith('}'):
                    json_str += '"}]}'  
                return json_str
            
            def trim_json(self, completion: str) -> str:
                client = self.load_model()  # Load the model like in `generate()`

                prompt = self.llm_fix_json_prompt(completion)

                # Assuming client has a method similar to `generate_text(prompt, None)`
                _, fixed_json = client.generate_text(prompt, None,False)

                fixed_json = fixed_json.strip()

                # Optional: Validate the output is valid JSON
                try:
                    json.loads(fixed_json)
                except json.JSONDecodeError as e:
                    logger.warning("Detected JSON truncation. Trying naive fix.")
                    fixed_json = self.fix_common_truncation(fixed_json)
                    try:
                        json.loads(fixed_json)
                    except json.JSONDecodeError as e2:
                        raise ValueError(f"Model returned invalid JSON (even after fix): {e2}\n\nReturned:\n{fixed_json}")
                return fixed_json





        self.llm = FloTorchLLMWrapper(evaluator_llm)
        self.async_config = AsyncConfig(run_async=async_run, max_concurrent=max_concurrent)
        self.custom_metrics = custom_metrics or []
        self.metric_args = metric_args 

        # Initialize DeepEval metrics from the registry
        DeepEvalEvaluationMetrics.initialize_metrics(llm=self.llm, metric_args=self.metric_args)

    def _build_test_cases(self, data: List[EvaluationItem]) -> List[LLMTestCase]:
        """
        Converts evaluation data into LLM test cases for DeepEval evaluation.
        """
        return [
            LLMTestCase(
                input=item.question,
                actual_output=item.generated_answer,
                expected_output=item.expected_answer,
                retrieval_context=item.context or [],
                context=item.context or []
            )
            for item in data
        ]

    def evaluate(
        self,
        data: List[EvaluationItem],
        metrics: Optional[List[str]] = None
    ) -> Dict[str, Any]:
        test_cases = self._build_test_cases(data)
        #example to fetch metrics, use like this
        if metrics is None:
            metrics = DeepEvalEvaluationMetrics.available_metrics()

        selected_metrics = [
            DeepEvalEvaluationMetrics.get_metric(m)
            for m in metrics
        ]
        eval_results = evaluate(
            test_cases=test_cases,
            async_config=self.async_config,
            metrics=selected_metrics + self.custom_metrics,
            error_config=ErrorConfig(ignore_errors=True)
        )
        return eval_results.model_dump()




from typing import Optional, Dict, Union, Mapping
from deepeval.metrics import (
    FaithfulnessMetric,
    ContextualRelevancyMetric,
    ContextualPrecisionMetric,
    ContextualRecallMetric,
    AnswerRelevancyMetric,
    HallucinationMetric,
)
from flotorch_core.evaluator.metrics.base_metrics import BaseEvaluationMetric
# from flotorch_core.evaluator.custom_metrics import CustomMetric
from flotorch_core.evaluator.metrics.metrics_keys import MetricKey

class DeepEvalEvaluationMetrics(BaseEvaluationMetric):
    _registry = {
        MetricKey.FAITHFULNESS: {
            "class": FaithfulnessMetric,
            "default_args": {"threshold": 0.7, "truths_extraction_limit": 30}
        },
        MetricKey.CONTEXT_RELEVANCY: {
            "class": ContextualRelevancyMetric,
            "default_args": {"threshold": 0.7}
        },
        MetricKey.CONTEXT_PRECISION: {
            "class": ContextualPrecisionMetric,
            "default_args": {"threshold": 0.7}
        },
        MetricKey.CONTEXT_RECALL: {
            "class": ContextualRecallMetric,
            "default_args": {"threshold": 0.7}
        },
        MetricKey.ANSWER_RELEVANCE: {
            "class": AnswerRelevancyMetric,
            "default_args": {"threshold": 0.7}
        },
        MetricKey.HALLUCINATION: {
            "class": HallucinationMetric,
            "default_args": {"threshold": 0.5}
        },
        # "custom_metric": {
        #     "class": CustomMetric,
        #     "default_args": {"threshold": 0.5}
        # },
    }

    _initialized_metrics: Dict[str, object] = {}

    @classmethod
    def available_metrics(cls) -> list[str]:
        return list(cls._initialized_metrics.keys())

    @classmethod
    def initialize_metrics(
        cls,
        llm,
        metric_args: Optional[
            Mapping[Union[str, MetricKey], Dict[str, Union[str, float, int]]]
        ] = None,
    ):
        """
        Initializes metric instances and stores them in an internal dictionary.

        This method iterates over the registered metrics in `cls._registry`, initializes each metric,
        and stores it in `cls._initialized_metrics`.  
        You can optionally override default arguments for each metric using `metric_args`.

        Args:
            llm: A language model wrapper instance (used as `model` argument for each metric class).
            metric_args (Optional): A dictionary providing argument overrides for each metric.
                The structure is:

                {
                    "metric_name": {
                        "arg1": value1,
                        "arg2": value2,
                        ...
                    },
                    ...
                }

                - The outer key corresponds to a metric name (as used in `cls._registry`).
                - The inner dictionary provides argument overrides that will replace defaults.

        Example Usage:
            metric_args={
                "faithfulness": {
                "threshold": 0.8
                }                    
            }
            
        """

        cls._initialized_metrics = {}
        metric_args = metric_args or {} 

        for name, config in cls._registry.items():
            args = config["default_args"].copy()
            args.update(metric_args.get(name, {}))  # override defaults
            cls._initialized_metrics[name] = config["class"](model=llm, **args)

    @classmethod
    def get_metric(cls, name: str) -> Dict[str, object]:
        """
        Retrieves an initialized metric instance by name.

        Args:
            name (str): The name of the metric to retrieve (must match a name in `cls._registry`).

        Returns:
            object: The initialized metric instance.

        Raises:
            ValueError: If the requested metric has not been initialized (i.e. not present in `cls._initialized_metrics`).

        Example Usage:
            faithfulness_metric = get_metric("faithfulness")
        """
        if name not in cls._initialized_metrics:
            raise ValueError(f"Metric '{name}' has not been initialized.")
        return cls._initialized_metrics[name]

    
        